cvs_id = "$Id: URLFetch.py,v 1.4 2003/02/20 12:56:07 juri Exp $"

import httplib_async
import asyncore
import urlparse
import base64
import gzip
try:
    from cStringIO import StringIO
except:
    from StringIO import StringIO
from error import log
import straw
import httplib_async
import time

class ConnectionManager:
    def __init__(self):
        self._queue = []

    def request(self, uri, consumer, headers={}, user=None, password=None,
                priority=straw.NetworkConstants.PRIORITY_DEFAULT):
        scheme, host, path, params, query, fragment = urlparse.urlparse(uri)
        assert scheme == "http", "only supports HTTP requests"
        try:
            host, port = host.split(":", 1)
            port = int(port)
        except (TypeError, ValueError):
            port = 80 # default port
        if not path:
            path = "/"
        if params:
            path = path + ";" + params
        if query:
            path = path + "?" + query
        straw.lookup_manager.lookup(host, self._request_resolved,
                                    (port, path, consumer, headers, user,
                                     password, priority))

    def _request_resolved(self, host, ip, data):
        if ip is not None:
            i = 0
            cell = (host, ip) + data
            while i < len(self._queue):
                if self._queue[i][8] > data[6]:
                    self._queue.insert(i, cell)
                    return
                i += 1
            self._queue.append(cell)
        else:
            data[2].http_failed("Host name lookup failed")

    def poll(self, timeout=0.1):
        # activate up to MAX_CONNECTIONS channels
        while self._queue and len(asyncore.socket_map) < straw.NetworkConstants.MAX_CONNECTIONS:
            self.do_request(*(self._queue.pop(0)[0:8]))
        # keep the network running
        now = time.time()
        straw.lookup_manager.poll(timeout)
        timeout -= (time.time() - now)
        if timeout > 0:
            asyncore.poll(timeout)
        # return non-zero if we should keep polling
        return len(self._queue) or len(asyncore.socket_map)

    def do_request(self, host, ip, port, path, consumer,
                   headers={}, user=None, password=None):
        return ProxyConsumer(host, ip, port, path, consumer, headers, user,
                             password)

connection_manager = ConnectionManager()

class Consumer:
    def __init__(self, host, ip, port):
        self.connection = httplib_async.HTTPConnection_async(host, ip, port, self)

    def request(self, path, headers={}, user=None, password=None):
        if user and password:
            headers['Authorization'] = 'Basic %s' % base64.encodestring(
                '%s:%s' % (user, password)).strip()
        self.connection.request("GET", path, headers)
        self.connection.execute()

    def finished_callback(self):
        raise NotImplementedException
        
    def http_failed(self, exception):
        """Called by HTTPConnection_async when connection failed with exception"""
        raise exception

    def http_header(self, status, header):
        """Called by HTTPConnection_async with status and header"""
        if header.getheader('content-length', '0') == '0':
            self.finished_callback()

    def http_redirect(self, location):
        """Called by HTTPConnection_async with the new location in case of 301 or 302"""
        raise NotImplementedException

    def feed(self, data):
        """Called by HTTPConnection_async with (part of the) data, after http_header"""
        raise NotImplementedException

class ProxyConsumer(Consumer):
    def __init__(self, host, ip, port, path, consumer,
                 headers={}, user=None, password=None):
        self.consumer = consumer
        self.data = ""
        self.header = headers
        self.status = None
        self.user = user
        self.password = password
        self.path = path
        Consumer.__init__(self, host, ip, port)
        headers['Accept-encoding'] = 'gzip'
        headers['User-agent'] = 'Straw %s' % straw.VERSION
        self.request(path, headers, user, password)
        self.start_time = time.time()

    def finished_callback(self):
        if self.header.getheader('content-encoding') == 'gzip':
            self.data = gzip.GzipFile(fileobj = StringIO(self.data)).read()
        self.consumer.http_results(self.status, self.header, self.data)

    def http_failed(self, exception):
        self.consumer.http_failed(exception)

    def http_header(self, status, header):
        self.status = status
        self.header = header
        Consumer.http_header(self, status, header)

    def http_redirect(self, location):
        connection_manager.request(location, self.consumer, self.header,
                                   self.user, self.password)

    def feed(self, data):
        self.data += data
        datalength = len(self.data)
        if datalength >= int(self.header.getheader('content-length', 0)):
            self.connection.close()
            self.finished_callback()
        elif datalength >= straw.NetworkConstants.MAX_DOWNLOAD_SIZE:
            self.connection.close()
            self.http_failed("Maximum download file size exceeded")
        elif time.time() - self.start_time > straw.NetworkConstants.MAX_DOWNLOAD_TIME:
            self.connection.close()
            self.http_failed("Maximum download time exceeded")
